// ----
// ---- file   : SyModule.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2009-2012 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 10May2009, ..., 08Jun2009, 
// ----          22May2010, 23May2010, 01Jul2010, 24Sep2010, 30Sep2010, 06Oct2010, 07Oct2010
// ----          10Oct2010, 17Oct2010, 23Jan2011, 05Feb2012
// ----
// ----
// ----

#define YAC_EPSILONCOMPARE_ABS defined
#define YAC_EPSILONCOMPARE_ABS_DEFAULT defined

#include "tksynth.h"

sSI SyModule::next_tick_count = 1;


sF32 SyModule::NoteToFreq(sF32 *_freqTable, sF32 _note) {
   if(_note >= 0)
   {
      if(_note >= 127)
      {
         _note = 126.9999f;
      }
      sUI intNote = (sUI) _note;
      sF32 frac = _note - (sF32)intNote;
      return _freqTable[intNote] * (1.0f - frac) + _freqTable[intNote + 1] * frac;
   }
   else
   {
      _note = - _note;
      if(_note >= 127)
      {
         _note = 126.9999f;
      }
      sUI intNote = (sUI) _note;
      sF32 frac = _note - (sF32)intNote;
      return - (_freqTable[intNote] * (1.0f - frac) + _freqTable[intNote + 1] * frac );
   }
}


SyModule::SyModule(void) {
   ////patch       = NULL;
   ui_instance_id  = -1;
   ui_position_x   = 0.0f;
   ui_position_y   = 0.0f;
   ui_layout_style = 0;

   inputs        = NULL;
   num_inputs    = 0;

   outputs        = NULL;
   num_outputs    = 0;

   clock_tick_count = -1;

   name = NULL;

   b_force_sampleexact_mode = 0;

   mix_rate = 8000; // should be overridden when instantiated;
}

SyModule::~SyModule() {
   freeInputs();
   freeOutputs();

   YAC_DELETE_SAFE(name);
}

void SyModule::setName(YAC_Object *_name) {
   if(NULL == name)
   {
      name = YAC_New_String();
   }
   name->yacCopy(_name);
}

YAC_Object *SyModule::getName(void) {
   return name;
}

void SyModule::freeInputs(void) {
   if(NULL != inputs)
   {
      sUI i=0;
      for(; i<num_inputs; i++)
      {
         SyInput *input = inputs[i];
         YAC_DELETE(input);
         inputs[i] = NULL;
      }
      delete [] inputs;
      inputs = NULL;

      num_inputs = 0;
   }
}

void SyModule::allocInputs(sUI _num) {
   inputs = new SyInput*[_num];

   for(sUI i=0; i<_num; i++)
   {
      SyInput *input = (SyInput*) YAC_NEW_POOLED(SyInput);
      input->module = this;
      inputs[i] = input;
      input->name = YAC_New_String();
   }

   // Note: derived modules should be careful not to add more inputs than allocated
}

SyInput *SyModule::addInput(const char *_name, sUI _mode, sF32 _minResetValue, sF32 _maxResetValue, sF32 _resetValue) {

   return initInput(num_inputs++, _name, _mode, _minResetValue, _maxResetValue, _resetValue);
}

SyInput *SyModule::initInput(sUI _idx, const char *_name, sUI _mode, sF32 _minResetValue, sF32 _maxResetValue, sF32 _resetValue) {
   // (note) used in SyModuleVST
   ////printf("xxx SyModule::initInput: inputs=0x%p num_inputs=%u\n", inputs, num_inputs);
   SyInput *input = inputs[_idx];

   ////printf("xxx SyModule::initInput: idx=%u/%u input=0x%p\n", _idx, num_inputs, inputs[_idx]);

   input->name->copy(_name);
   input->reset_value     = _resetValue;
   input->min_reset_value = _minResetValue;
   input->max_reset_value = _maxResetValue;
   input->mode            = _mode;
   input->default_mode    = _mode;

   return input;
}

void SyModule::freeOutputs(void) {
   if(NULL != outputs)
   {
      sUI i=0;
      for(; i<num_outputs; i++)
      {
         SyOutput *output = outputs[i];
         YAC_DELETE(output);
         outputs[i] = NULL;
      }
      delete [] outputs;
      outputs = NULL;

      num_outputs = 0;
   }
}

void SyModule::allocOutputs(sUI _num) {
   outputs = new SyOutput*[_num];
   for(sUI i=0; i<_num; i++)
   {
      SyOutput *output = (SyOutput*) YAC_NEW_POOLED(SyOutput);
      output->module = this;
      output->name = YAC_New_String();
      outputs[i] = output;
   }
   // Note: derived modules should be careful not to add more outputs than allocated
}

SyOutput *SyModule::addOutput(const char *_name, sUI _mode) {
   return initOutput(num_outputs++, _name, _mode);
}

SyOutput *SyModule::initOutput(sUI _idx, const char *_name, sUI _mode) {
   // (note) used in SyModuleVST

   ////printf("xxx SyModule::initOutput: outputs=0x%p num_outputs=%u\n", outputs, num_outputs);

   ///printf("xxx SyModule::initOutput: idx=%u/%u output=0x%p\n", _idx, num_outputs, outputs[_idx]);

   SyOutput *output = outputs[_idx];

   output->name->copy(_name);
   output->mode         = _mode;
   output->default_mode = _mode;

   //printf("xxx num_outputs=%d output=%p\n", num_outputs, output);

   return output;
}

void SyModule::init(void) {
   // cleanup 
   //freeInputs();
   //freeOutputs();

   // derived class should alloc/add inputs/outputs here
}

void SyModule::setUiInstanceId(sSI _id) {
   ui_instance_id = _id;
}

sSI SyModule::getUiInstanceId(void) {
   return ui_instance_id;
}

void SyModule::setUiPositionX(sF32 _x) {
   ui_position_x = _x;
}

sF32 SyModule::getUiPositionX(void) {
   return ui_position_x;
}

void SyModule::setUiPositionY(sF32 _y) {
   ui_position_y = _y;
}

sF32 SyModule::getUiPositionY(void) {
   return ui_position_y;
}

void SyModule::setUiLayoutStyle(sSI _style) {
   ui_layout_style = _style;
}

sSI SyModule::getUiLayoutStyle(void) {
   return ui_layout_style;
}

sUI SyModule::getNumInputs(void) {
   return num_inputs;
}

YAC_Object *SyModule::getInputById(sUI _id) {
   if(_id < num_inputs)
   {
      return inputs[_id];
   }
   return NULL;
}

YAC_Object *SyModule::getInputByName(YAC_String *_name) {
   sUI i;
   for(i=0; i<num_inputs; i++)
   {
      SyInput *input = inputs[i];
      if(input->name->yacEquals(_name))
      {
         return input;
      }
   }
   return NULL;
}

sSI SyModule::getInputIdByName(YAC_String *_name) {
   sUI i;
   for(i=0; i<num_inputs; i++)
   {
      SyInput *input = inputs[i];
      if(input->name->yacEquals(_name))
      {
         return (sSI) i;
      }
   }
   return -1;
}

sUI SyModule::getNumOutputs(void) {
   return num_outputs;
}

YAC_Object *SyModule::getOutputById(sUI _id) {
   if(_id < num_outputs)
   {
      return outputs[_id];
   }
   return NULL;
}

YAC_Object *SyModule::getOutputByName(YAC_String *_name) {
   sUI i;
   for(i=0; i<num_outputs; i++)
   {
      SyOutput *output = outputs[i];
      if(output->name->yacEquals(_name))
      {
         return output;
      }
   }
   return NULL;
}

sSI SyModule::getOutputIdByName(YAC_String *_name) {
   sUI i;
   for(i=0; i<num_outputs; i++)
   {
      SyOutput *output = outputs[i];
      if(output->name->yacEquals(_name))
      {
         return (sSI) i;
      }
   }
   return -1;
}

void SyModule::reset(void) {

   clock_tick_count = -2;

   // Send reset signal to all input connections
   sUI i;
   for(i=0; i<num_inputs; i++)
   {
      SyInput *input = inputs[i];
      if(input->connections)
      {
         SyConnection *conn = input->connections;
         while(conn)
         {
            ////printf("xxx SyModule::reset: conn.tc=%d\n", conn->module->clock_tick_count);
            if(conn->module->clock_tick_count != -2)
            {
               conn->module->reset();
            }
            conn = conn->next;
         }
      }
   }

   // Reset input values
   for(i=0; i<num_inputs; i++)
   {
      SyInput *input = inputs[i];

      sF32 resetVal = input->reset_value * input->level + input->offset;

      // Do reset for feedback mode
      Dsyloopallch()
      {
         input->last_value[chIdx]    = resetVal;
         input->current_value[chIdx] = resetVal;
      }

      // Do reset for non-feedback mode
      sUI k;
      for(k=0; k< (SYPORT_MAXCHANNELS * SYPORT_MAXBUFSIZE); k++)
      {
         input->buffer[k] = resetVal;
      }

      ////printf("xxx SyModule::reset: input[%d].reset_value=%f\n", i, input->reset_value);
   }

   // Reset output values
   for(i=0; i<num_outputs; i++)
   {
      SyOutput *output = outputs[i];

      // Do reset for feedback mode
      Dsyloopallch()
      {
         Dsyout(output, 0.0f);
      }

      // Do reset for non-feedback mode
      sUI k;
      for(k=0; k< (SYPORT_MAXCHANNELS * SYPORT_MAXBUFSIZE); k++)
      {
         output->buffer[k] = 0.0f;
      }
   }
}

void SyModule::calc(sSI _tickCount, sUI _framesAdvance) {
   sUI i;

   ////printf("xxx calc tickCount=%d class=%s\n", _tickCount, yacClassName());

   // Update tick count
   clock_tick_count = _tickCount;

   //printf("xxx SyModule::calc: tickCount=%d this=%p num_inputs=%d\n", _tickCount, this, num_inputs);

   sUI numInputs = getNumInputs(); // may differ from num_inputs (e.g. SyModuleIO)

   // Remember last input value (for interpolation)
   for(i=0; i<numInputs; i++)
   {
      SyInput *input = inputs[i];
      Dsyloopch(input)
      {
         input->last_value[chIdx] = Dsyval(input);
      }
   }

   // Calc/sum input connection values and determine effective input values
   for(i=0; i<numInputs; i++)
   {
      SyInput *input = inputs[i];

      // Any outputs connected to this input ?
      if(input->connections)
      {
         // Iterate current input channels
         Dsyloopch(input)
         {
            sF32 cval = 0.0f;

            // Iterate current input connections
            SyConnection *conn = input->connections;
            while(conn)
            {
               if(conn->module->clock_tick_count != _tickCount)
               {
                  conn->module->calc(_tickCount, _framesAdvance);
               }
               SyOutput *connOutput = conn->module->outputs[conn->output_id];
               if(2 == connOutput->num_channels)
               {
                  // Stereo output to mono or stereo input
                  cval += connOutput->current_value[chIdx] * conn->level + conn->offset;
               }
               else
               {
                  // Mono output to mono or stereo input
                  cval += connOutput->current_value[0] * conn->level + conn->offset;
               }
               conn = conn->next;
            }

           
            switch(input->clamp_mode)
            {
               default:
               case SYINPUT_CLAMP_NONE:
               case SYINPUT_CLAMP_TRIG:
                  break;
                  
               case SYINPUT_CLAMP_CLIP:
               {
                  if(cval > input->max_reset_value)
                  {
                     cval = input->max_reset_value;
                  }
                  else if(cval < input->min_reset_value)
                  {
                     cval = input->min_reset_value;
                  }
               }
               break;
               
               case SYINPUT_CLAMP_TANH:
               {
                  // Normalize to -n..n range, 0=min, 1=max value
                  sF32 r = (input->max_reset_value - input->min_reset_value);
                  if(r > 0.0f)
                  {
                     cval = (cval - input->min_reset_value) / r;
                     cval = 0.5f * tanh( (cval-0.5f) * 2.0f) + 0.5f;
                     cval = Dyac_denorm_32(cval);
                  }
                  else
                  {
                     cval = 0.0f;
                  }
                  // Rescale to original range
                  cval = (cval * r) + input->min_reset_value;
               }
               break;

               case SYINPUT_CLAMP_TANH2: // actually this is a bug but can sound quite nice. "test_distort.st2" uses this.
               {
                  // Normalize to -n..n range, 0=min, 1=max value
                  sF32 r = (input->max_reset_value - input->min_reset_value);
                  if(r > 0.0f)
                  {
                     cval = (cval - input->min_reset_value) / r;
                  }
                  else
                  {
                     cval = 0.0f;
                  }
                  cval = tanh(cval);
                  // Rescale to original range
                  cval = (cval * r) + input->min_reset_value;
               }
               break;

            }

            // eff. input val= sum of connections * totalLevel + totalOffset
            cval = 
               (input->last_value[chIdx] * (1.0f - input->interpol_amount)) + 
               (cval * input->level + input->offset) * input->interpol_amount
               ;

            // Store final value
            input->current_value[chIdx] = cval;

             // if(input->clamp_mode)
             // {
             //    printf("xxx clamp: cval=%f minReset=%f maxReset=%f\n", cval, input->max_reset_value, input->max_reset_value);
             // }

         } // iterate input ports

      } // if connections
      else
      {
         // No connections, copy reset value
         Dsyloopch(input)
         {
            input->current_value[chIdx] = input->reset_value * input->level + input->offset;
         }
      }

   } // iterate inputs

   // Now calc output values of this module (implemented in derived classes)
   sUI numOutputs = getNumOutputs();
   if(numOutputs > 0)
   {
      doCalc(_framesAdvance);
   }
}

void SyModule::calcBlock(sSI _tickCount, sUI _numFrames) {
   sUI i;

   ////printf("xxx SyModule::calcBlock: this=%p tickCount=%d numFrames=%u num_inputs=%d\n", this, _tickCount, _numFrames, num_inputs);

   // Update tick count
   clock_tick_count = _tickCount;

   // Calc/sum input connection values and determine effective input values
   for(i=0; i<num_inputs; i++)
   {
      SyInput *input = inputs[i];

      if(input->connections)
      {
         // Zero input buffer
         sF32 *inbuf = input->buffer;
         memset(inbuf, 0, _numFrames * input->num_channels * sizeof(sF32));

         sBool bSEMode = ((SYPORT_MODE_SE == input->mode)||input->b_force_sampleexact_mode);

         const sUI inbufNumSamples = (  
            bSEMode ? (_numFrames * input->num_channels) : input->num_channels
                                        );
         sUI j, k;

         // Iterate current input connections
         SyConnection *conn = input->connections;
         while(conn)
         {
            if(conn->module->clock_tick_count != _tickCount)
            {
               conn->module->calcBlock(_tickCount, _numFrames);
            }

            // Get connection output port
            SyOutput *connOutput = conn->module->outputs[conn->output_id];
            sF32 *connbuf = connOutput->buffer;
            sF32 connLevel  = conn->level;
            const sF32 connOffset = conn->offset;

            // Add connection output to input buffer

            ////printf("xxx SyModule::calcBlock: BEGIN add connection output id %d to input %d\n", conn->output_id, i);

            if(bSEMode)
            {

               // Sample-exact mode (#frames samples per block)
               if(input->mode == connOutput->mode)
               {
                  // No mode conversion needed

                  if(input->num_channels == connOutput->num_channels)
                  {
                     // No channel conversion needed
                     if(Dfltequal(connLevel, 1.0f) && Dfltequal(connOffset, 0.0f))
                     {
                        for(k=0; k<inbufNumSamples; k++)
                        {
                           inbuf[k] += connbuf[k];
                        }
                     }
                     else
                     {
                        for(k=0; k<inbufNumSamples; k++)
                        {
                           inbuf[k] += connbuf[k] * connLevel + connOffset;
                        }
                     }
                  }
                  else if((1 == input->num_channels) && (2 == connOutput->num_channels))
                  {
                     // Convert stereo to mono
                     connLevel *= 0.5f;
                     j = 0;
                     for(k=0; k<inbufNumSamples; k++, j+=2)
                     {
                        inbuf[k] += (connbuf[j]+connbuf[j+1]) * connLevel + connOffset;
                     }
                  }
                  else
                  {
                     // Convert mono to stereo
                     j = 0;
                     for(k=0; k<inbufNumSamples; k+=2, j++)
                     {
                        sF32 t = connbuf[j] * connLevel + connOffset;
                        inbuf[k+0] += t;
                        inbuf[k+1] += t;
                     }
                  }
               }
               else
               {
                  // Convert block-exact connection values to sample exact input buffer (repeat values)

                  if(input->num_channels == connOutput->num_channels)
                  {
                     // No channel conversion needed
                     if(1 == input->num_channels)
                     {
                        // Mono
                        sF32 t = connbuf[0] * connLevel + connOffset;

                        for(k=0; k<inbufNumSamples; k++)
                        {
                           inbuf[k] += t;
                        }
                     }
                     else
                     {
                        // Stereo
                        sF32 l = connbuf[0] * connLevel + connOffset;
                        sF32 r = connbuf[1] * connLevel + connOffset;

                        for(k=0; k<_numFrames; k+=2)
                        {
                           inbuf[k+0] += l;
                           inbuf[k+1] += r;
                        }
                     }
                  }
                  else if((1 == input->num_channels) && (2 == connOutput->num_channels))
                  {
                     // Convert stereo to mono
                     sF32 t = (connbuf[0] + connbuf[1]) * 0.5f * connLevel + connOffset;

                     for(k=0; k<inbufNumSamples; k++)
                     {
                        inbuf[k] += t;
                     }
                  }
                  else
                  {
                     // Convert mono to stereo
                     sF32 t = connbuf[0] * connLevel + connOffset;

                     for(k=0; k<_numFrames; k+=2)
                     {
                        inbuf[k+0] += t;
                        inbuf[k+1] += t;
                     }
                  }
               }
            }
            else
            {
               // Block-exact mode (one sample per block)

               if(input->mode == connOutput->mode)
               {
                  // No mode conversion needed

                  if(input->num_channels == connOutput->num_channels)
                  {
                     // No channel conversion needed

                     if(1 == input->num_channels)
                     {
                        // Mono
                        inbuf[0] += connbuf[0] * connLevel + connOffset;
                     }
                     else
                     {
                        // Stereo
                        inbuf[0] += connbuf[0] * connLevel + connOffset;
                        inbuf[1] += connbuf[1] * connLevel + connOffset;
                     }
                  }
                  else if((1 == input->num_channels) && (2 == connOutput->num_channels))
                  {
                     // Convert stereo to mono
                     inbuf[0] += (connbuf[0] + connbuf[1]) * 0.5f * connLevel + connOffset;
                  }
                  else
                  {
                     // Convert mono to stereo
                     sF32 t = connbuf[0] * connLevel + connOffset;
                     inbuf[0] += t;
                     inbuf[1] += t;
                  }
               }
               else
               {
                  // Convert sample-exact connection values to block exact input buffer (discard values)

                  if(input->num_channels == connOutput->num_channels)
                  {
                     // No channel conversion needed
                     for(k=0; k<input->num_channels; k++)
                     {
                        inbuf[k] += connbuf[k] * connLevel + connOffset;
                     }
                  }
                  else if((1 == input->num_channels) && (2 == connOutput->num_channels))
                  {
                     // Convert stereo to mono
                     sF32 t = (connbuf[0] + connbuf[1]) * 0.5f * connLevel + connOffset;
                     inbuf[0] += t;
                  }
                  else
                  {
                     // Convert mono to stereo
                     sF32 t = connbuf[0] * connLevel + connOffset;
                     inbuf[0] += t;
                     inbuf[1] += t;
                  }
               }

            } // check input port mode

            ////printf("xxx SyModule::calcBlock: END add connection output id %d to input %d\n", conn->output_id, i);

            // Next connection
            conn = conn->next;

         } // loop connections


         // Now apply input port post processing / clipping
         
         switch(input->clamp_mode)
         {
            default:
            case SYINPUT_CLAMP_NONE:
            case SYINPUT_CLAMP_TRIG:
               break;

            case SYINPUT_CLAMP_CLIP:
            {
               for(k=0; k<inbufNumSamples; k++)
               {
                  if(inbuf[k] > input->max_reset_value)
                  {
                     inbuf[k] = input->max_reset_value;
                  }
                  else if(inbuf[k] < input->min_reset_value)
                  {
                     inbuf[k] = input->min_reset_value;
                  }
               }
            }
            break;
            
            case SYINPUT_CLAMP_TANH:
            {
               // Normalize to -n..n range, 0=min, 1=max value
               sF32 r = (input->max_reset_value - input->min_reset_value);
               if(r > 0.0f)
               {
                  sF32 obr = 1.0f / r;
                  for(k=0; k<inbufNumSamples; k++)
                  {
                     sF32 t = inbuf[k];
                     t = (t - input->min_reset_value) * obr;
                     t = 0.5f * tanh( (t - 0.5f) * 2.0f) + 0.5f;
                     t = Dyac_denorm_32(t);
                     // Rescale to original range
                     inbuf[k] = (t * r) + input->min_reset_value;
                  }
               }
               else
               {
                  // Special case / misconfiguration
                  for(k=0; k<inbufNumSamples; k++)
                  {
                     inbuf[k] = input->min_reset_value;
                  }
               }
            }
            break;

            case SYINPUT_CLAMP_TANH2: // actually this is a bug but can sound quite nice. "test_distort.st2" uses this.
            {
               // Normalize to -n..n range, 0=min, 1=max value
               sF32 r = (input->max_reset_value - input->min_reset_value);
               if(r > 0.0f)
               {
                  sF32 obr = 1.0f / r;
                  for(k=0; k<inbufNumSamples; k++)
                  {
                     sF32 t = inbuf[k];
                     t = (t - input->min_reset_value) * obr;
                     t = tanh(t);
                     // Rescale to original range
                     inbuf[k] = (t * r) + input->min_reset_value;
                  }
               }
               else
               {
                  // Special case / misconfiguration
                  for(k=0; k<inbufNumSamples; k++)
                  {
                     inbuf[k] = input->min_reset_value;
                  }
               }
            }
            break;

         } // switch clamp mode


         // Apply interpolation ?
         if(Dfltnotequal(input->interpol_amount, 1.0f))
         {
            // Interpolate and scale to final range
            sF32 lastValue[SYPORT_MAXCHANNELS];
            for(k=0; k<input->num_channels; k++)
            {
               lastValue[k] = input->last_value[k];
            }

            for(k=0; k<inbufNumSamples; )
            {
               for(j=0; j<input->num_channels; j++)
               { 
                  inbuf[k] = 
                     (lastValue[j] * (1.0f - input->interpol_amount)) + 
                     (inbuf[k] * input->level + input->offset) * input->interpol_amount
                     ;
                  lastValue[j] = inbuf[k];
                  k++;
               }
            }

            // Remember for next calcBlock() call
            for(j=0; j<input->num_channels; j++)
            {
               input->last_value[j] = lastValue[j];
            }
         }
         else
         {
            // Simply scale to range
            for(k=0; k<inbufNumSamples; k++)
            {
               inbuf[k] = inbuf[k] * input->level + input->offset;
            }
         }

      } // if connections
      else
      {
         // No connections, use reset value (already in buffer)
      }

   } // iterate inputs

   // Now calc output buffers of this module (either implemented in derived classes or
   //  fallback code in this class)
   sUI numOutputs = getNumOutputs();
   if(numOutputs > 0)
   {
      if(b_force_sampleexact_mode)
      {
         SyModule::doCalcBlock(_numFrames);
      }
      else
      {
         doCalcBlock(_numFrames);
      }
   }
}

sBool SyModule::saveStream(YAC_Object *_ofs) {
   // Write version
   _ofs->yacStreamWriteI16(0x0000);

   return 1;
}

sBool SyModule::readStream(YAC_Object *_ifs) {
   // Read version
   sU16 v = _ifs->yacStreamReadI16();
   if(0 != v)
   {
      printf("[---] SyModule::readStream: default implementation called but version(%d) != 0.\n", v);
      return 0;
   }
   return 1;
}

void SyModule::doCalc(sUI _framesAdvance) {
   printf("[~~~] SyModule::doCalc: NO-OP default implementation called! (class=\"%s\")\n", yacClassName());
}

void SyModule::doCalcBlock(sUI _numFrames) {
   //
   // Default / fallback implementation for modules that do not support calcBlockMode()
   //

   //printf("xxx SyModule::doCalcBlock: fallback. #frames=%u\n", _numFrames);

   sUI seTickCount = 0x12345678;

   sUI i;
   for(i=0; i<_numFrames; i++)
   {
      sUI numInputs = getNumInputs();
      
      // Copy from input buffer to input value
      sUI inIdx;
      for(inIdx=0; inIdx<numInputs; inIdx++)
      {
         SyInput *in = inputs[inIdx];
         sBool bSEMode = ((SYPORT_MODE_SE == in->mode) || in->b_force_sampleexact_mode);

         if(bSEMode)
         {
            // Sample-exact mode
            sUI off = i * in->num_channels;
            for(sUI chIdx=0; chIdx<in->num_channels; chIdx++)
            {
               in->current_value[chIdx] = in->buffer[off++];
            }
         }
         else
         {
            // Block mode, repeat first sample
            for(sUI chIdx=0; chIdx<in->num_channels; chIdx++)
            {
               in->current_value[chIdx] = in->buffer[chIdx];
            }
         }
      }

      doCalc(1);
      
      // Copy output current_value(s) to output buffer(s)

      sUI numOutputs = getNumOutputs();

      sUI outIdx;
      for(outIdx=0; outIdx<numOutputs; outIdx++)
      {
         SyOutput *out = outputs[outIdx];

         sUI chIdx;
         sUI off = i * out->num_channels;
         for(chIdx=0; chIdx<out->num_channels; chIdx++)
         {
            out->buffer[off++] = out->current_value[chIdx];
         }
      } 

      //printf("xxx doCalcBlock: fallback. i=%u\n", i);

      seTickCount++;
   }

}

void SyModule::setEnableForceSampleExactMode(sSI _bEnabled) {
   b_force_sampleexact_mode = _bEnabled;
}

sSI SyModule::getEnableForceSampleExactMode(void) {
   return b_force_sampleexact_mode;
}

void SyModule::setMixRate(sF32 _mixRate) {
   if(_mixRate < 1000.0f)
   {
      mix_rate = 1000.0f;
   }
   else if(_mixRate > 192000)
   {
      mix_rate = 192000;
   }
   else
   {
      mix_rate = _mixRate;
   }
}

sF32 SyModule::getMixRate(void) {
   return mix_rate;
}

void SyModule::setTickDuration(sF32 _tickDuration) {
   (void)_tickDuration;
}

