// ----
// ---- file   : SyModuleDlyModulate.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2010-2013 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 15Jun2010, 30Sep2010, 07Oct2010, 23Jan2011, 28May2013
// ----          
// ----
// ----
// ----

#include "tksynth.h"

#include "SyModuleDlyModulate.h"


SyModuleDlyModulate::SyModuleDlyModulate(void) {
   state = new SyModuleDlyModulateState[SYPORT_MAXCHANNELS];
}

SyModuleDlyModulate::~SyModuleDlyModulate() {
   delete [] state;
   state = NULL;
}


void SyModuleDlyModulate::init(void) {
   SyModule::init();

   allocInputs(9);
   addInput("smp",          SYPORT_MODE_SE, -1.0f,      1.0f,   1.0f); // #0
   addInput("amount",       SYPORT_MODE_BE,  0.0f,      1.0f,   1.0f); // #1
   addInput("millisec",     SYPORT_MODE_BE,  0.0f,   1000.0f, 250.0f); // #2
   addInput("feedback",     SYPORT_MODE_BE,  0.0f,      1.0f,   0.6f); // #3
   addInput("depth",        SYPORT_MODE_BE,  0.0f,      1.0f,   0.0f); // #4
   addInput("speed",        SYPORT_MODE_BE,  0.01f, 16000.0f,   5.0f); // #5 (Hz)
   addInput("modamount",    SYPORT_MODE_BE,  0.0f,      1.0f,   0.0f); // #6
   addInput("dlyamount",    SYPORT_MODE_BE,  0.0f,      1.0f,   0.5f); // #7
   addInput("modfbkamount", SYPORT_MODE_BE,  0.0f,      1.0f,   0.0f); // #8

   allocOutputs(1);
   addOutput("smp", SYPORT_MODE_SE);  // #0
}

void SyModuleDlyModulate::reset(void) {
   SyModule::reset();

   memset(state, 0, sizeof(SyModuleDlyModulateState) * SYPORT_MAXCHANNELS);
   ringbuf_offset = 0;

   angle = 0.0f;
}

void SyModuleDlyModulate::doCalc(sUI _framesAdvance) {

   SyInput *in_smp          = inputs[0];
   SyInput *in_amount       = inputs[1];
   SyInput *in_millisec     = inputs[2];
   SyInput *in_feedback     = inputs[3];
   SyInput *in_depth        = inputs[4];
   SyInput *in_speed        = inputs[5];
   SyInput *in_modamount    = inputs[6];
   SyInput *in_dlyamount    = inputs[7];
   SyInput *in_modfbkamount = inputs[8];

   SyOutput *output = outputs[0];

   // // sSI dlyNumSamples = (sSI) TKSYNTH_MILLISEC_ADV_TO_FRAMES(Dsyctl(in_millisec), _framesAdvance);

   sSI dlyNumSamples = (sSI) TKSYNTH_MILLISEC_TO_FRAMES(Dsyctl(in_millisec));

   if(dlyNumSamples < 1)
   {
      dlyNumSamples = 1;
   }
   else if(dlyNumSamples > SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES)
   {
      dlyNumSamples = SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES;
   }

   // Calculate offset to delayed sample
   sSI dlyOff = ringbuf_offset - dlyNumSamples;

   // Calculate offset to delay-modulated sample
   sF32 depth = Dsyctl(in_depth);
   sF32 dlyShift = (sF32) (depth * dlyNumSamples * (0.5f+0.5f*cos(angle))); // Vibrato
   sF32 dlyOffModFrac = dlyShift - (sSI)dlyShift;
   sF32 dlyOffMod = ringbuf_offset - 1 - dlyShift;

   sSI dlyOffModC = (sSI) dlyOffMod;
   sSI dlyOffModN = (sSI) dlyOffMod + 1;
   // // sF32 dlyOffModFrac = dlyOffMod - (sF32)dlyOffModC;
   // printf("xxx frac=%f\n", dlyOffModFrac);

   // Current sample integer offset
   while(dlyOff < 0)
      dlyOff += SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES;
   while(dlyOff > (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1))
      dlyOff -= (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1);

   // Current modulated sample integer offset
   while(dlyOffModC < 0)
      dlyOffModC += SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES;
   while(dlyOffModC > (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1))
      dlyOffModC -= (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1);

   // Next modulated sample integer offset
   while(dlyOffModN < 0)
      dlyOffModN += SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES;
   while(dlyOffModN > (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1))
      dlyOffModN -= (SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES-1);

   //printf("xxx ringOff=%d depth=%f dlyOffMod=%f modOff=%f\n", ringbuf_offset, depth, dlyOffMod, (1 + dlyShift));

   // Dry/wet amount
   sF32 amount = Dsyctl(in_amount);
   if(amount < 0.0f)
      amount = 0.0f;
   else if(amount > 1.0f)
      amount = 1.0f;

   // Determine feedback amount
   sF32 feedback = Dsyctl(in_feedback);
   if(feedback < 0.0f)
      feedback = 0.0f;
   else if(feedback > 1.0f)
      feedback = 1.0f;

   // Determine modulation amount
   sF32 modamount = Dsyctl(in_modamount);
   if(modamount < 0.0f)
      modamount = 0.0f;
   else if(modamount > 1.0f)
      modamount = 1.0f;

   // Determine delay amount
   sF32 dlyamount = Dsyctl(in_dlyamount);
   if(dlyamount < 0.0f)
      dlyamount = 0.0f;
   else if(dlyamount > 1.0f)
      dlyamount = 1.0f;

   // Determine dry/wet mod feedback amount
   sF32 modfbkamount = Dsyctl(in_modfbkamount);
   if(modfbkamount < 0.0f)
      modfbkamount = 0.0f;
   else if(modfbkamount > 1.0f)
      modfbkamount = 1.0f;

   // Calculate output samples
   Dsyloopch(in_smp)
   {
      Dsystate(SyModuleDlyModulateState);
      
      // Get current sample input
      sF32 smp = Dsyval(in_smp);

      // Get delayed sample
      sF32 dlyVal = st->ringbuf[dlyOff];

      // Get delay-modulated sample (linear interpolation)
      sF32 dlyValMod = st->ringbuf[dlyOffModC] + (st->ringbuf[dlyOffModN] - st->ringbuf[dlyOffModC]) * dlyOffModFrac;

      sF32 out = smp + (dlyVal - smp) * amount;
      sF32 outmod = smp + ( (dlyVal * dlyamount + dlyValMod * modamount) - smp) * amount;

      Dsyout(output, outmod);

      // Store next sample in ringbuffer
      st->ringbuf[ringbuf_offset] = smp + (out * feedback * (1.0f - modfbkamount) + outmod * feedback * modfbkamount);
   }

   // Rotate ringbuffer
   ringbuf_offset++;
   if(ringbuf_offset >= SYMODULEDLYMODULATE_MAX_DELAY_SAMPLES)
   {
      ringbuf_offset = 0;
   }

   // Modulate LFO
   sF32 spd = Dsyctl(in_speed);
   if(spd <= 0.0f)
      spd = 0.0001f;
   else if(spd >= 16000.0f)
      spd = 16000.0f;
   angle += (sF32) ((2.0f*sM_PI * spd * _framesAdvance) / mix_rate);

   while(angle >= (sF32) (2.0f*sM_PI))
   {
      angle -= (sF32) (2.0f*sM_PI);
   }
}
