// ----
// ---- file   : SyModuleDlySimple.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2010-2013 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 15Jun2010, 30Sep2010, 07Oct2010, 05Feb2012, 28May2013
// ----          
// ----
// ----
// ----

#include "tksynth.h"

#include "SyModuleDlySimple.h"


SyModuleDlySimple::SyModuleDlySimple(void) {
   state = new SyModuleDlySimpleState[SYPORT_MAXCHANNELS];
}

SyModuleDlySimple::~SyModuleDlySimple() {
   delete [] state;
   state = NULL;
}


void SyModuleDlySimple::init(void) {
   SyModule::init();

   allocInputs(9);
   addInput("smp",       SYPORT_MODE_SE, -1.0f,    1.0f,   1.0f); // #0
   addInput("amount",    SYPORT_MODE_BE,  0.0f,    1.0f,   1.0f); // #1
   addInput("millisec",  SYPORT_MODE_BE,  0.0f, 1000.0f, 250.0f); // #2
   addInput("feedback",  SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f); // #3
   addInput("timeshift", SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f); // #4
   addInput("sync",      SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f); // #5
   addInput("dampen",    SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f); // #6
   addInput("freeze",    SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f); // #7
   addInput("speed",     SYPORT_MODE_BE, -4.0f,    4.0f,   1.0f); // #8

   allocOutputs(1);
   addOutput("smp", SYPORT_MODE_SE);  // #0
}

void SyModuleDlySimple::reset(void) {
   SyModule::reset();

   memset(state, 0, sizeof(SyModuleDlySimpleState) * SYPORT_MAXCHANNELS);
   ringbuf_offset = 0.0f;

   dampen_framecount = 0;
}

void SyModuleDlySimple::setTickDuration(sF32 _tickDuration) {
   tick_duration = _tickDuration;
}

void SyModuleDlySimple::doCalc(sUI _framesAdvance) {

   SyInput *in_smp        = inputs[0];
   SyInput *in_amount     = inputs[1];
   SyInput *in_millisec   = inputs[2];
   SyInput *in_feedback   = inputs[3];
   SyInput *in_timeshift  = inputs[4];
   SyInput *in_sync       = inputs[5];
   SyInput *in_dampen     = inputs[6];
   SyInput *in_freeze     = inputs[7];
   SyInput *in_speed      = inputs[8];

   SyOutput *output = outputs[0];

   sBool bFreeze = (Dsyctl(in_freeze) >= 0.5f);

   sF32 milliSec = Dsyctl(in_millisec);

   if(Dsyctl(in_sync) >= 0.5f)
   {
      // Synchronize to replay ticks
      milliSec = milliSec * tick_duration;
   }

   // // sSI dlyNumSamples = TKSYNTH_MILLISEC_ADV_TO_FRAMES(milliSec, _framesAdvance);
   sSI dlyNumSamples = (sSI) TKSYNTH_MILLISEC_TO_FRAMES(Dsyctl(in_millisec));

   if(dlyNumSamples < 1)
   {
      dlyNumSamples = 1;
   }
   else if(dlyNumSamples > SYMODULEDLYSIMPLE_MAX_DELAY_SAMPLES)
   {
      dlyNumSamples = SYMODULEDLYSIMPLE_MAX_DELAY_SAMPLES;
   }

   sSI ringbufOffset = (sSI) ringbuf_offset;

   // Calculate sample shift
   sSI dlyShift = (sSI) (Dsyctl(in_timeshift) * dlyNumSamples);
   if(dlyShift < 0)
      dlyShift = 0;
   else if(dlyShift > dlyNumSamples)
      dlyShift = dlyNumSamples;

   // Calculate offset to delayed sample
   sSI dlyOff = ringbufOffset - dlyNumSamples + 1 + dlyShift;
   while(dlyOff < 0)
      dlyOff += dlyNumSamples;
   while(dlyOff > (dlyNumSamples-1))
      dlyOff = (dlyNumSamples-1);

   // Dry/wet amount
   sF32 amount = Dsyctl(in_amount);
   if(amount < 0.0f)
      amount = 0.0f;
   else if(amount > 1.0f)
      amount = 1.0f;

   // Calculate output samples
   Dsyloopch(in_smp)
   {
      Dsystate(SyModuleDlySimpleState);

      sF32 smp;
      if(!bFreeze)
      {
         smp= Dsyval(in_smp);
      }
      else
      {
         smp = 0.0f;
      }
      sF32 dlyVal = st->ringbuf[dlyOff];

      sF32 out = smp + (dlyVal - smp) * amount;

      st->next_smp     = smp;
      st->last_dly_smp = out;

      Dsyout(output, out);
   }

   // Determine feedback amount
   sF32 feedback = Dsyctl(in_feedback);
   if(feedback < 0.0f)
      feedback = 0.0f;
   else if(feedback > 1.0f)
      feedback = 1.0f;

   // Filter delay buffer
   dampen_framecount += _framesAdvance;
   while(dampen_framecount >= 2048)//(dlyNumSamples)
   {
      sF32 dampen = Dsyctl(in_dampen);
      if(dampen > 0.0f)
      {
         if(dampen > 1.0f)
            dampen = 1.0f;
         else if(dampen < 0.0f)
            dampen = 0.0f;

         Dsyloopch(in_smp)
         {
            Dsystate(SyModuleDlySimpleState);
            
            sSI i;
            sF32 l = st->ringbuf[dlyNumSamples-1];
            for(i=0; i<=dlyNumSamples; i++)
            {
               sF32 c = st->ringbuf[i];
               sF32 n = c - (c-l) * dampen;
               st->ringbuf[i] = Dyac_denorm_32(n);
               l = c;
            }
         }

         // Next input channel
      }

      dampen_framecount -= 2048;//dlyNumSamples;
   }

   if(!bFreeze)
   {
      // Store next sample(s) in ringbuffer(s)
      Dsyloopch(in_smp)
      {
         Dsystate(SyModuleDlySimpleState);
         
         st->push(ringbufOffset, feedback);
      }
   }

   // Rotate ringbuffer
   sF32 ringSpeed = Dsyctl(in_speed);
   if(ringSpeed != 0.0f)
      ringSpeed = 1.0f / ringSpeed;

   if(ringSpeed > 32.0f)
      ringSpeed = 32.0f;

   ringbuf_offset += ringSpeed;

   while(ringbuf_offset >= (sF32)dlyNumSamples)
   {
      ringbuf_offset -= (sF32)dlyNumSamples;
   }

   while(ringbuf_offset < 0)
   {
      ringbuf_offset += (sF32)dlyNumSamples;
   }

   ////output->current_value = (sF32) (sin(angle + in_phase->current_value) * output->level * in_volume->current_value + output->offset);
}
