// ----
// ---- file   : SyModuleDynCompressor.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2010-2013 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 29Jun2010, 30Jun2010, 30Sep2010, 07Oct2010, 23Jan2011, 28May2013
// ----
// ----
// ----

#include "tksynth.h"

#include "SyModuleDynCompressor.h"


float SyModuleDynCompressor::rms_weights[SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE];


void SyModuleDynCompressor::init(void) {
   SyModule::init();

   allocInputs(7);
   addInput("smp",        SYPORT_MODE_SE, -1.0f,     1.0f, 0.0f); // #0 
   addInput("smpctl",     SYPORT_MODE_SE, -1.0f,     1.0f, 0.0f); // #1 side chain
   addInput("threshold",  SYPORT_MODE_BE,  0.001f,   1.0f, 0.5f); // #2 env trigger threshold
   addInput("ratio",      SYPORT_MODE_BE,  1.0f,   100.0f, 4.0f); // #3 ratio
   addInput("attackspd",  SYPORT_MODE_BE,  0.1f,   500.0f, 2.0f); // #4 attack speed (millisec)
   addInput("releasespd", SYPORT_MODE_BE,  0.1f,   500.0f, 2.0f); // #5 release speed (millisec)
   addInput("amount",     SYPORT_MODE_BE,  0.0f,     1.0f, 1.0f); // #6 dry/wet amount

   allocOutputs(3);
   addOutput("smp",  SYPORT_MODE_SE);
   addOutput("env",  SYPORT_MODE_SE);
   addOutput("gain", SYPORT_MODE_SE);

   CalcRMSWeights(); // (todo) only needs to be done for the first compressor instance

   memset(buf, 0, sizeof(float) * SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE);
   buf_offset = 0;

   env_mode = 0;
}

void SyModuleDynCompressor::CalcRMSWeights(void) {
   sSI i;
   sF32 a = 0.0f;
   sF32 w = (sF32) (sM_PI * (1.0f/SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE));
   sF32 sum = 0.0f;
   for(i=0; i<SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE; i++)
   {
      sF32 c = sin(a);
      rms_weights[i] = c;
      sum += c;
      a += w;
   }
   sF32 nscl = (SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE / sum);
   for(i=0; i<SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE; i++)
   {
      rms_weights[i] *= nscl;
   }
}

sF32 SyModuleDynCompressor::calcRMS(sF32 _inSmp) {
   // push incoming sample into ring buffer, apply window fxn and return real mean square

   buf[buf_offset] = _inSmp;
   buf_offset++; // rotate ringbuf
   if(buf_offset >= SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE)
   {
      buf_offset -= SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE;
   }

   sF32 avg = 0.0f;
   sSI off = buf_offset - 1;
   sSI ci = 0;
   for(sSI i=0; i<SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE; i++)
   {
      if(off < 0)
      {
         off += SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE;
      }
      sF32 smp = buf[off--];
      avg += smp * smp * rms_weights[ci++];
   }
   return (sF32) sqrt(avg / SYMODULE_DYN_COMPRESSOR_RMS_WINDOW_SIZE);
}

void SyModuleDynCompressor::reset(void) {
   SyModule::reset();

   env_mode = 0;
   env_phase = 0.0f;
}

void SyModuleDynCompressor::doCalc(sUI _framesAdvance) {

   SyInput *in_smp        = inputs[0];
   SyInput *in_smpctl     = inputs[1];
   SyInput *in_threshold  = inputs[2];
   SyInput *in_ratio      = inputs[3];
   SyInput *in_attackspd  = inputs[4];
   SyInput *in_releasespd = inputs[5];
   SyInput *in_amount     = inputs[6];

   SyOutput *out_smp  = outputs[0];
   SyOutput *out_env  = outputs[1];
   SyOutput *out_gain = outputs[2];

   sF32 attackSpd  = Dsyclamp(Dsyctl(in_attackspd),  0.1f, 500.0f);
   sF32 releaseSpd = Dsyclamp(Dsyctl(in_releasespd), 0.1f, 500.0f);
   sF32 ratio      = Dsyclamp(Dsyctl(in_ratio),      1.0f, 100.0f);
   sF32 amount     = Dsyclamp(Dsyctl(in_amount),     0.0f,   1.0f);

   attackSpd  = 1000.0f / (attackSpd  * mix_rate);
   releaseSpd = 1000.0f / (releaseSpd * mix_rate);

   // Calculate average of side-chain input channels
   sF32 inSmp = 0.0f;
   Dsyloopch(in_smpctl)
   {
      inSmp += Dsyval(in_smpctl);
   }
   inSmp *= (1.0f / in_smpctl->num_channels);

   sF32 rmsSmp = calcRMS(inSmp);

   sF32 threshold = Dsyclamp(Dsyctl(in_threshold), 0.01f, 1.0f);
   sBool bTrigger = (rmsSmp > threshold);
   if(bTrigger)
   {
      if(0 == env_mode)
      {
         // Trigger attack envelope
         env_phase = 0.0f;
         env_mode  = 1;
         damp      = 1.0f / ((rmsSmp / threshold) * ratio);
         printf("xxx comp rmsSmp=%f ratio=%f damp=%f\n", rmsSmp, ratio, damp);
      }
   }

   sF32 gain;


   switch(env_mode)
   {
      default:
      case 0: // idle
         gain = 1.0f;
         break;

      case 1: // attack
         ////gain = (1.0f - env_phase) * damp;
         gain = (1.0f - env_phase) + (env_phase * damp);
         env_phase += attackSpd;
         if(env_phase > 1.0f)
         {
            // Hold attack envelope until not triggered any longer
            if(bTrigger)
            {
               env_phase = 1.0f;
            }
            else
            {
               // Start release phase
               env_mode  = 2;
               env_phase = 0.0f;
            }
         }
         break;

      case 2: // release
         //gain = env_phase * damp;
         gain = env_phase + ((1.0f - env_phase) * damp);
         ////printf("xxx comp env_phase=%f gain=%f\n", env_phase, gain);
         env_phase += releaseSpd;
         if(env_phase > 1.0f)
         {
            // End of release envelope
            env_phase = 0.0f;
            env_mode  = 0;
         }
         break;
   }

   // Apply dynamic gain to input samples
   Dsyloopch(in_smp)
   {
      sF32 smp = Dsyval(in_smp);
      sF32 out = smp * gain;
      out = (amount * out) + (1.0f - amount) * smp;
      Dsyout(out_smp, out);
   }

   // Output tracked envelope
   sUI chIdx = 0;
   Dsyout(out_env, rmsSmp);

   // Output compressor gain
   Dsyout(out_gain, gain);
   
}
