// ----
// ---- file   : SyModuleEQ3.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2009-2011 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 10May2009, ..., 08Jun2090, 
// ----          22May2010, 30Sep2010, 07Oct2010, 12Oct2010, 23Jan2011
// ----
// ----
// ----

#include "tksynth.h"

#include "SyModuleEQ3.h"

// see <http://musicdsp.org/showArchiveComment.php?ArchiveID=236>


//#define XXX_DEBUG defined

#ifdef XXX_DEBUG
static sSI xxx_smpcount = 0;
#define XXX_SMPCOUNT_IVAL ((1<<15)-1)
#endif // XXX_DEBUG



static double vsa = (1.0 / 4294967295.0); // Very small amount (Denormal Fix)


void SyModuleEQ3::init(void) {
   SyModule::init();

   allocInputs(6);

   addInput("smp",     SYPORT_MODE_SE,   -1.0f,     1.0f,    0.0f); // #0
   addInput("freqlo",  SYPORT_MODE_BE,    5.0f,  2000.0f,  880.0f); // #1
   addInput("gainlo",  SYPORT_MODE_BE,    0.0f,     4.0f,    1.0f); // #2
   addInput("freqhi",  SYPORT_MODE_BE, 2000.0f, 12000.0f, 5000.0f); // #3
   addInput("gainhi",  SYPORT_MODE_BE,    0.0f,     4.0f,    1.0f); // #4
   addInput("gainmid", SYPORT_MODE_BE,    0.0f,     4.0f,    1.0f); // #5


   allocOutputs(1);

   addOutput("smp", SYPORT_MODE_SE); // #0

   memset(state, 0, 2*sizeof(SyModuleEQ3State));
}


void SyModuleEQ3::reset(void) {
   SyModule::reset();

   Dsyloopallch()
   {
      Dsystate(SyModuleEQ3State);
      st->reset();
   }
}

void SyModuleEQ3::doCalc(sUI _framesAdvance) {

   /* compute result */
   SyInput *in_smp       = inputs[0];

   SyInput *in_freq1     = inputs[1];
   SyInput *in_gain1     = inputs[2];
   SyInput *in_freq2     = inputs[3];
   SyInput *in_gain2     = inputs[4];
   SyInput *in_gain3     = inputs[5];

   SyOutput *output = outputs[0];

   // Cutoff/center frequency
   sF32 freq1 = Dsyctl(in_freq1);
   if(freq1 < 20.0f)
      freq1 = 20.0f;
   if(freq1 > 1000.0f)
      freq1 = 1000.0f;

   sF32 freq2 = Dsyctl(in_freq2);
   if(freq2 < 1100.0f)
      freq2 = 1100.0f;
   if(freq2 > 16000.0f)
      freq2 = 16000.0f;


   // Calculate filter cutoff frequencies
   sF64 lf = 2 * sin(sM_PI * freq1 / mix_rate);
   sF64 hf = 2 * sin(sM_PI * freq2 / mix_rate);

   // Filter sample
   Dsyloopch(in_smp)
   {
      Dsystate(SyModuleEQ3State);

      sF64 sample = Dsyval(in_smp);

      // Filter #1 (lowpass)
      st->f1p0 += (lf * (sample - st->f1p0)) + vsa;
      st->f1p1 += (lf * (st->f1p0 - st->f1p1));
      st->f1p2 += (lf * (st->f1p1 - st->f1p2));
      st->f1p3 += (lf * (st->f1p2 - st->f1p3));

      sF64 l = st->f1p3;

      // Filter #2 (highpass)
      st->f2p0 += (hf * (sample - st->f2p0)) + vsa;
      st->f2p1 += (hf * (st->f2p0 - st->f2p1));
      st->f2p2 += (hf * (st->f2p1 - st->f2p2));
      st->f2p3 += (hf * (st->f2p2 - st->f2p3));

      sF64 h = st->sdm3 - st->f2p3;

      // Calculate midrange (signal - (low + high))
      sF64 m = st->sdm3 - (h + l);

      // Scale and Combine
      l *= Dsyctl(in_gain1);
      m *= Dsyctl(in_gain3);
      h *= Dsyctl(in_gain2);
   
      sF64 out = (l + m + h);

      Dsyout(output, out);
   }

   // Shuffle history buffer
   Dsyloopch(in_smp)
   {
      Dsystate(SyModuleEQ3State);

      st->sdm3 = st->sdm2;
      st->sdm2 = st->sdm1;
      st->sdm1 = Dsyval(in_smp);
   }
}

void SyModuleEQ3::doCalcBlock(sUI _numFrames) {

   // (note) faster than the fallback code but not really optimized

   /* compute result */
   SyInput *in_smp       = inputs[0];

   SyOutput *output = outputs[0];

   if(in_smp->num_channels != output->num_channels)
   {
      return SyModule::doCalcBlock(_numFrames);
   }

   SyInput *in_freq1     = inputs[1];
   SyInput *in_gain1     = inputs[2];
   SyInput *in_freq2     = inputs[3];
   SyInput *in_gain2     = inputs[4];
   SyInput *in_gain3     = inputs[5];


   // Cutoff/center frequency
   sF32 freq1 = Dsyctlbm(in_freq1);
   if(freq1 < 20.0f)
      freq1 = 20.0f;
   if(freq1 > 1000.0f)
      freq1 = 1000.0f;

   sF32 freq2 = Dsyctlbm(in_freq2);
   if(freq2 < 1100.0f)
      freq2 = 1100.0f;
   if(freq2 > 16000.0f)
      freq2 = 16000.0f;


   // Calculate filter cutoff frequencies
   sF64 lf = 2 * sin(sM_PI * freq1 / mix_rate);
   sF64 hf = 2 * sin(sM_PI * freq2 / mix_rate);

   const sF32 gain1 = Dsyctlbm(in_gain1);
   const sF32 gain3 = Dsyctlbm(in_gain3);
   const sF32 gain2 = Dsyctlbm(in_gain2);

   // Filter samples
   sUI k = 0;
   for(sUI i=0; i<_numFrames; i++)
   {
      Dsyloopch(in_smp)
      {
         Dsystate(SyModuleEQ3State);

         sF64 sample = in_smp->buffer[k];
         
         // Filter #1 (lowpass)
         st->f1p0 += (lf * (sample - st->f1p0)) + vsa;
         st->f1p1 += (lf * (st->f1p0 - st->f1p1));
         st->f1p2 += (lf * (st->f1p1 - st->f1p2));
         st->f1p3 += (lf * (st->f1p2 - st->f1p3));

         sF64 l = st->f1p3;

         // Filter #2 (highpass)
         st->f2p0 += (hf * (sample - st->f2p0)) + vsa;
         st->f2p1 += (hf * (st->f2p0 - st->f2p1));
         st->f2p2 += (hf * (st->f2p1 - st->f2p2));
         st->f2p3 += (hf * (st->f2p2 - st->f2p3));

         sF64 h = st->sdm3 - st->f2p3;

         // Calculate midrange (signal - (low + high))
         sF64 m = st->sdm3 - (h + l);

         // Scale and Combine
         l *= gain1;
         m *= gain3;
         h *= gain2;
   
         sF64 out = (l + m + h);

         output->buffer[k] = (sF32) (out * output->level + output->offset);

         // Shuffle history buffer
         st->sdm3 = st->sdm2;
         st->sdm2 = st->sdm1;
         st->sdm1 = sample;

         k++;
      }
   }

}
