// ----
// ---- file   : SyModuleFltPhaser6St.cpp
// ---- author : (c) 1998 by Ross Bencina <rossb@kagi.com>
// ----          (c) 2010-2011 by Bastian Spiegel <bs@tkscript.de>
// ---- legal  : 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 24Jun2010, 30Sep2010, 07Oct2010, 23Jan2011
// ----          
// ----
// ----
// ----

/*
    class: Phaser
    implemented by: Ross Bencina <rossb@kagi.com>
    date: 24/8/98

    Phaser is a six stage phase shifter, intended to reproduce the
    sound of a traditional analogue phaser effect.
    This implementation uses six first order all-pass filters in
    series, with delay time modulated by a sinusoidal.

    This implementation was created to be clear, not efficient.
    Obvious modifications include using a table lookup for the lfo,
    not updating the filter delay times every sample, and not
    tuning all of the filters to the same delay time.

    Thanks to:
    The nice folks on the music-dsp mailing list, including...
    Chris Towsend and Marc Lindahl
    
    ...and Scott Lehman's Phase Shifting page at harmony central:
    http://www.harmony-central.com/Effects/Articles/Phase_Shifting/

*/


#include "tksynth.h"

#include "SyModuleFltPhaser6St.h"


void SyModuleFltPhaser6St::init(void) {
   SyModule::init();

   allocInputs(6);
   addInput("smp",      SYPORT_MODE_SE, -1.0f,     1.0f,    0.0f); // #0
   addInput("freqlo",   SYPORT_MODE_BE,  0.1f, 16000.0f,  400.0f); // #1
   addInput("freqhi",   SYPORT_MODE_BE,  0.1f, 16000.0f, 1000.0f); // #2
   addInput("feedback", SYPORT_MODE_BE,  0.0f,     1.0f,    0.3f); // #3
   addInput("rate",     SYPORT_MODE_BE,  0.0f, 16000.0f,    3.0f); // #4
   addInput("depth",    SYPORT_MODE_BE,  0.0f,     1.0f,    0.3f); // #5

   allocOutputs(1);
   addOutput("smp", SYPORT_MODE_SE); // #0

   memset(state, 0, sizeof(SyModuleFltPhaser6StState) * SYPORT_MAXCHANNELS);

   lfo_phase = 0.0f;
}

void SyModuleFltPhaser6St::reset(void) {
   SyModule::reset();
}

void SyModuleFltPhaser6St::doCalc(sUI _framesAdvance) {

   SyInput *in_smp      = inputs[0];
   SyInput *in_freqlo   = inputs[1];
   SyInput *in_freqhi   = inputs[2];
   SyInput *in_feedback = inputs[3];
   SyInput *in_rate     = inputs[4];
   SyInput *in_depth    = inputs[5];

   SyOutput *out_smp = outputs[0];

   sF32 freqLo   =         Dsyclamp(Dsyctl(in_freqlo),   0.1f, 16000.0f) / (mix_rate*0.5f);
   sF32 freqHi   =         Dsyclamp(Dsyctl(in_freqhi),   0.1f, 16000.0f) / (mix_rate*0.5f);
   sF32 feedback =         Dsyclamp(Dsyctl(in_feedback), 0.0f,     1.0f);
   sF32 rate     = (sF32)((Dsyclamp(Dsyctl(in_rate),     0.1f, 16000.0f) / (mix_rate)) * _framesAdvance * 2.0f * sM_PI);
   sF32 depth    =         Dsyclamp(Dsyctl(in_depth),    0.0f,     1.0f);

   // Calculate and update phaser sweep lfo
   float d  = freqLo + (freqHi - freqLo) * ((sin(lfo_phase) + 1.0f) * 0.5f);
   
   lfo_phase = Dsywrap(lfo_phase + rate, 0, (sF32)(2.0f * sM_PI));

   Dsyloopch(in_smp)
   {
      Dsystate(SyModuleFltPhaser6StState);

      // Update filter coeffs
      for(int i=0; i<6; i++ )
      {
        	st->alps[i].delay(d);
      }

      sF32 inSmp = Dsyval(in_smp);

      // Calculate output
      float out = st->alps[0].update(
         st->alps[1].update(
            st->alps[2].update(
               st->alps[3].update(
                  st->alps[4].update(
                     st->alps[5].update(inSmp + st->zm1 * feedback)
                                     )
                                  )
                               )
                            )
                                     );
      st->zm1 = out;

      out = inSmp + out * depth;

      Dsyout(out_smp, out);
   }
}
