// ----
// ---- file   : SyModuleOscSawDSFBlit.cpp
// ---- author : Bastian Spiegel <bs@tkscript.de>
// ---- legal  : (c) 2009-2011 by Bastian Spiegel. 
// ----          Distributed under terms of the GNU LESSER GENERAL PUBLIC LICENSE (LGPL). See 
// ----          http://www.gnu.org/licenses/licenses.html#LGPL or COPYING for further information.
// ----
// ---- info   : This is part of the "syntracker" midi/audio sequencer.
// ----
// ---- changed: 10May2009, ..., 08Jun2090, 
// ----          22May2010, 02Jul2010, 30Sep2010, 07Oct2010, 23Jan2011
// ----
// ----
// ----

#include "tksynth.h"

#include "SyModuleOscSawDSFBLIT.h"

// see <http://musicdsp.org/showArchiveComment.php?ArchiveID=90>



/* Bandlimited synthesis of sawtooth by
 * leaky integration of a DSF BLIT
 *
 * Emanuel Landeholm, March 2002
 * emanuel.landeholm@telia.com
 *
 * Provided \"as is\".
 * Free as in Ef Are Ee Ee.
 */





/* initializes a lowpass, sets cutoff/leakiness */
void SyModuleOscSawDSFBLIT::initLowpass(sF64 cutoff)
{
  sF64 Omega;

  blit.lowpass.x1 = blit.lowpass.y1 = 0.0;

  Omega = atan(sM_PI * cutoff);
  blit.lowpass.a = -(1.0 - Omega) / (1.0 + Omega);
  blit.lowpass.b =  (1.0 - blit.lowpass.b) / 2.0;
}


/* initializes the blit structure
 *
 * The aNQ parameter is the desired attenuation
 * at nyquist. A low value yields a duller
 * sawtooth but gets rid of those annoying CLICKS
 * when sweeping the frequency up real high. |aNQ|
 * must be strictly less than 1.0! Find a setting
 * which works for you.
 *
 * The cutoff parameter controls the leakiness of
 * the integrator.
 */
void SyModuleOscSawDSFBLIT::initBlit(sF64 aNQ, sF64 cutoff)
{
   blit.phase  = 0.0;
   blit.aNQ = aNQ;
   blit.curcps = 0.0;
   blit.curper = 0.0;
   blit.val = 0.0;
   initLowpass(cutoff);
}


sF64 SyModuleOscSawDSFBLIT::updateBlit(sF64 _cps, sF64 _phase)
{
   sF64 P2, beta, Nbeta, cosbeta, n, d, saw;
  
   ////if(blit.phase >= 1.0 || blit.curcps == 0.0)
   {
      /* New cycle, update frequency and everything
       * that depends on it
       */
      blit.curcps = _cps;        /* this cycle\'s frequency */
      blit.curper = 1.0 / _cps;  /* this cycle\'s period */
      
      P2 = blit.curper / 2.0;
      blit.N = 1.0 + floor(P2); /* # of partials incl. dc */
      
      /* find the roll-off parameter which gives
       * the desired attenuation at nyquist
       */
      blit.a   = pow(blit.aNQ, 1.0 / P2);
      blit.aN  = pow(blit.a,   blit.N);
   }


   // Note:
   //  the oscillator becomes unstable if the phase modulation rate exceeds ~1hz
   //  in other words: the phase input should only be used
   //  for constants or slow controller changes
  
   sF32 blitPhase = (sF32) (blit.phase + _phase);
   while(blitPhase >= 1.0)
      blitPhase -= 1.0;
   while(blitPhase < 0.0)
      blitPhase += 1.0;


   beta = (2*sM_PI) * blitPhase;
   
   Nbeta = blit.N * beta;
   cosbeta = cos(beta);
   
   /* The dsf blit is scaled by 1 / period to give approximately the same
    * peak-to-peak over a wide range of frequencies.
    */
   
   n = 1.0 -
      blit.aN * cos(Nbeta) -
      blit.a * (cosbeta - blit.aN * cos(Nbeta - beta));
   d = blit.curper * (1.0 + blit.a * (-2.0 * cosbeta + blit.a));
   
  
   blit.val = n / d - blit.curcps; /* This division can only fail if |a| == 1.0
                              * Subtracting the fundamental frq rids of DC
     */
   
   // update low pass
   /* shape blit spectrum into a saw */
   blit.lowpass.newy = blit.lowpass.b * (blit.val + blit.lowpass.x1) - blit.lowpass.a * blit.lowpass.y1;

   saw = blit.lowpass.newy; 
   
   return saw;
}


void SyModuleOscSawDSFBLIT::init(void) {
   SyModule::init();

   allocInputs(4);
   addInput("volume", SYPORT_MODE_BE, -4.0f,    4.0f,   1.0f);  // #0
   addInput("pitch",  SYPORT_MODE_BE, 10.0f, 2000.0f, 440.0f);  // #1
   addInput("phase",  SYPORT_MODE_BE,  0.0f,    1.0f,   0.0f);  // #2
   addInput("sync",   SYPORT_MODE_SE,  0.0f,    1.0f,   0.0f);  // #3

   allocOutputs(2);
   addOutput("smp",  SYPORT_MODE_SE); // #0
   addOutput("sync", SYPORT_MODE_SE); // #1

}

void SyModuleOscSawDSFBLIT::reset(void) {
   SyModule::reset();

   memset(&blit, 0, sizeof(blit));
   initBlit(0.5, 0.0001);
}

void SyModuleOscSawDSFBLIT::doCalc(sUI _framesAdvance) {

   SyInput *in_volume = inputs[0];
   SyInput *in_pitch  = inputs[1];
   SyInput *in_phase  = inputs[2];
   SyInput *in_sync   = inputs[3];

   SyOutput *output   = outputs[0];
   SyOutput *out_sync = outputs[1];

   // Oscillator sync
   sF32 sync = Dsyctl(in_sync);
   if(b_synczerox)
   {
      if(Dsyfltsign(sync) ^ Dsyfltsign(Dsylastctl(in_sync)))
      {
         initBlit(0.5, 0.0001);
      }
   }
   else
   {
      if(sync > 0.0f)
      {
         initBlit(0.5, 0.0001);
      }
   }

   // Get pitch
   sF64 pitch = mapOscPitch(Dsyctl(in_pitch));
   if(pitch < 10)
   {
      pitch = 10;
   }
   else if(pitch > (0.5 * mix_rate))
   {
      pitch = 0.5 * mix_rate;
   }

   sF64 out = -4.0 * updateBlit(pitch / mix_rate, Dsyctl(in_phase));

   out *= Dsyctl(in_volume);

   Dsyoutmono(output, out);
   ////output->current_value = (sF32) (out * output->level * in_volume->current_value + output->offset);


   // tick
   blit.phase += blit.curcps * _framesAdvance; /* update phase */

   while(blit.phase >= 1.0)
   {
      blit.phase -= 1.0;
      sync = 1.0f;
   }
   while(blit.phase < 0.0)
   {
      blit.phase += 1.0;
      sync = 1.0f;
   }

   // Cycle
   blit.lowpass.x1 = blit.val;
   blit.lowpass.y1 = blit.lowpass.newy;

   Dsyoutmono(out_sync, sync);
}

